/*
 * The SmartConfig Library
 * Copyright (C) 2004-2006
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * For further informations on the SmartConfig Library please visit
 *
 *                        http://smartconfig.sourceforge.net
 */
package net.smartlab.config;

import java.net.URI;
import java.net.URISyntaxException;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.StringTokenizer;

import org.xml.sax.Attributes;

/**
 * This class provides a skeletal implementation of a configuration tree to
 * provide a set of common operations with a generic implementation.
 * 
 * @author rlogiacco
 * @uml.dependency supplier="net.smartlab.config.ConfigurationException"
 */
abstract public class Configuration extends Node {

	/**
	 * @uml.property name="resolve"
	 */
	private boolean resolve;

	/**
	 * @uml.property name="listeners"
	 */
	protected Collection listeners = new ArrayList();

	/**
	 * @uml.property name="monitor"
	 */
	protected Monitor monitor = new Monitor();


	/**
	 * Creates an instance of the class as an <code>Element</code> without a
	 * parent node. This constructor is providen for subclasses convenience.
	 */
	protected Configuration() {
		super(null, null, null);
		this.monitor.start();
	}

	/**
	 * Initializes this configuration with a name and a set of attributes.
	 * 
	 * @param name the name of the configuration tree.
	 * @param attributes a set of attributes relative to the global
	 *        configuration tree.
	 */
	protected void init(String name, Attributes attributes) {
		this.name = name;
		if (attributes != null) {
			this.attributes = new HashMap(attributes.getLength());
			for (int i = 0; i < attributes.getLength(); i++) {
				this.attributes.put(attributes.getQName(i), attributes.getValue(i));
			}
		}
		this.parent = this;
	}

	/**
	 * @see net.smartlab.config.Node#resolve(java.lang.String, java.lang.String)
	 */
	protected Node resolve(String name, String id) throws ConfigurationException {
		Iterator children = this.children.iterator();
		while (children.hasNext()) {
			Element child = (Element)children.next();
			if (child instanceof Node && name.equals(child.name) && id.equals(child.getId())) {
				return (Node)child;
			}
		}
		throw new ConfigurationException("Cannot resolve element " + name + " with id " + id);
	}

	/**
	 * This method checks if the source from the configuration structure was
	 * modified since it was last read.
	 * 
	 * @return <code>true</code> if the underlying representation of the
	 *         structure was modified since last read.
	 * @exception ConfigurationException if something wrong occurred while
	 *            checking the underlaying structure like the file is no more
	 *            readable or the stream no more available.
	 */
	public abstract boolean isChanged() throws ConfigurationException;

	/**
	 * Ensures the in memory representation of the configuration matches with
	 * the underlying representation.
	 * 
	 * @exception ConfigurationException if the underlying representation of the
	 *            structure is no more accessible or the decoding process fails
	 *            due to an unappropriately initialized <code>Cipher</code>.
	 */
	public abstract void update() throws ConfigurationException;

	/**
	 * Sets the resolution method to adopt while querying this configuration
	 * tree. When set to <code>true</code> all internal references are
	 * automatically resolved to the destination node if present in the
	 * configuration tree.
	 * 
	 * @param resolve the resolution method to adopt.
	 * @uml.property name="resolve"
	 */
	public void setResolve(boolean resolve) {
		this.resolve = resolve;
	}

	/**
	 * Returns the resolution method adopted by this configuration tree.
	 * 
	 * @return <code>true</code> if internal references should be
	 *         automatically resolved.
	 * @uml.property name="resolve"
	 */
	public boolean getResolve() {
		return resolve;
	}

	/**
	 * Returns the registered changes listeners.
	 * 
	 * @return the registered changes listeners.
	 * @uml.property name="listeners"
	 */
	public Collection getListeners() {
		return Collections.unmodifiableCollection(listeners);
	}

	/**
	 * Adds a configuration changes listener.
	 * 
	 * @param listener the listener to add.
	 */
	public void addListener(Listener listener) {
		if (listener != null) {
			listeners.add(listener);
		}
	}

	/**
	 * Removes a configuration changes listener.
	 * 
	 * @param listener the listener to remove.
	 */
	public void removeListener(Listener listener) {
		listeners.remove(listener);
	}

	/**
	 * Returns an instance of a <code>Configuration</code> subclass
	 * accordingly to the provided URI.
	 * 
	 * Monitoring interval can be set through the system property
	 * <code>smartconfig.monitor.delay</code> expressed in milliseconds and
	 * defaulting to <b>60000</b>.
	 * 
	 * @param uri the URI to access the configuration.
	 * @return an instance of a <code>Configuration</code> subclass.
	 * @throws ConfigurationException if no handler matches the specified URI or
	 *         an error is encountered trying to access the configuration URI.
	 */
	public final static Configuration getInstance(String uri) throws ConfigurationException, URISyntaxException {
		return Configuration.getInstance(new URI(uri), (Collection)null);
	}

	/**
	 * Returns an instance of a <code>Configuration</code> subclass
	 * accordingly to the provided URI monitoring it for updates which will be
	 * notified to the provided listener.
	 * 
	 * Monitoring interval can be set through the system property
	 * <code>smartconfig.monitor.delay</code> expressed in milliseconds and
	 * defaulting to <b>60000</b>.
	 * 
	 * @param uri the URI to access the configuration.
	 * @param listener a <i>Listener</i> interface implementation.
	 * @return an instance of a <code>Configuration</code> subclass.
	 * @throws ConfigurationException if no handler matches the specified URI or
	 *         an error is encountered trying to access the configuration URI.
	 */
	public final static Configuration getInstance(String uri, Listener listener) throws ConfigurationException,
			URISyntaxException {
		return Configuration.getInstance(new URI(uri), Collections.singletonList(listener));
	}

	/**
	 * Returns an instance of a <code>Configuration</code> subclass
	 * accordingly to the provided URI monitoring it for updates which will be
	 * notified to the provided collection of listeners.
	 * 
	 * Monitoring interval can be set through the system property
	 * <code>smartconfig.monitor.delay</code> expressed in milliseconds and
	 * defaulting to <b>60000</b>.
	 * 
	 * @param uri the URI to access the configuration.
	 * @param listeners a collection of <i>Listener</i> interface implementors.
	 * @return an instance of a <code>Configuration</code> subclass.
	 * @throws ConfigurationException if no handler matches the specified URI or
	 *         an error is encountered trying to access the configuration URI.
	 */
	public final static Configuration getInstance(String uri, Collection listeners) throws ConfigurationException,
			URISyntaxException {
		return Configuration.getInstance(new URI(uri), listeners);
	}

	/**
	 * Returns an instance of a <code>Configuration</code> subclass
	 * accordingly to the provided URI.
	 * 
	 * Monitoring interval can be set through the system property
	 * <code>smartconfig.monitor.delay</code> expressed in milliseconds and
	 * defaulting to <b>60000</b>.
	 * 
	 * @param uri the URI to access the configuration.
	 * @return an instance of a <code>Configuration</code> subclass.
	 * @throws ConfigurationException if no handler matches the specified URI or
	 *         an error is encountered trying to access the configuration URI.
	 */
	public final static Configuration getInstance(URI uri) throws ConfigurationException {
		return Configuration.getInstance(uri, (Collection)null);
	}

	/**
	 * Returns an instance of a <code>Configuration</code> subclass
	 * accordingly to the provided URI monitoring it for updates which will be
	 * notified to the provided listener.
	 * 
	 * Monitoring interval can be set through the system property
	 * <code>smartconfig.monitor.delay</code> expressed in milliseconds and
	 * defaulting to <b>60000</b>.
	 * 
	 * @param uri the URI to access the configuration.
	 * @param listener a <i>Listener</i> interface implementation.
	 * @return an instance of a <code>Configuration</code> subclass.
	 * @throws ConfigurationException if no handler matches the specified URI or
	 *         an error is encountered trying to access the configuration URI.
	 */
	public final static Configuration getInstance(URI uri, Listener listener) throws ConfigurationException {
		return Configuration.getInstance(uri, Collections.singletonList(listener));
	}

	/**
	 * Returns an instance of a <code>Configuration</code> subclass
	 * accordingly to the provided URI monitoring it for updates which will be
	 * notified to the provided collection of listeners.
	 * 
	 * Monitoring interval can be set through the system property
	 * <code>smartconfig.monitor.delay</code> expressed in milliseconds and
	 * defaulting to <b>60000</b>.
	 * 
	 * @param uri the URI to access the configuration.
	 * @param listeners a collection of <i>Listener</i> interface implementors.
	 * @return an instance of a <code>Configuration</code> subclass.
	 * @throws ConfigurationException if no handler matches the specified URI or
	 *         an error is encountered trying to access the configuration URI.
	 */
	public final static Configuration getInstance(URI uri, Collection listeners) throws ConfigurationException {
		Configuration result = null;
		if (uri.getScheme().equals("jdbc")) {
			try {
				StringTokenizer tokenizer = new StringTokenizer(uri.toString(), "&=");
				while (tokenizer.hasMoreTokens()) {
					String token = tokenizer.nextToken();
					if (token.equalsIgnoreCase("driver")) {
						Class.forName(tokenizer.nextToken());
					}
				}
				result = new SQLConfiguration(DriverManager.getConnection(uri.toString()));
			} catch (SQLException sqle) {
				throw new ConfigurationException(sqle);
			} catch (ClassNotFoundException cnfe) {
				throw new ConfigurationException(cnfe);
			}
		} else if (uri.getScheme().equals("file") || uri.getScheme().startsWith("http")
				|| uri.getScheme().equals("ftp")) {
			try {
				result = new XMLConfiguration(uri.toURL());
			} catch (java.net.MalformedURLException murle) {
				throw new ConfigurationException(murle);
			}
		} else {
			throw new ConfigurationException("No available handler for " + uri.getScheme());
		}
		if (listeners != null) {
			result.listeners.addAll(listeners);
		}
		return result;
	}

	/**
	 * Checks if the underling configuration has been changed and notifies
	 * listeners of the change.
	 */
	protected final void check() {
		try {
			if (this.isChanged()) {
				this.update();
				Iterator notifications = this.listeners.iterator();
				while (notifications.hasNext()) {
					((Listener)notifications.next()).onUpdate(this);
				}
			}
		} catch (ConfigurationException ce) {
			// FIXME should we do something different like notify the listeners?
			ce.printStackTrace();
		}
	}


	/**
	 * Monitors the configuration for updates notifying changes to the
	 * registered listeners.
	 * 
	 * @author rlogiacco
	 */
	private class Monitor extends Thread {

		/**
		 * Delay, in milliseconds, between configuration update checks.
		 */
		private long sleep = Long.parseLong(System.getProperty("smartconfig.monitor.delay", "60000"));


		/**
		 * @see java.lang.Thread#run()
		 */
		public void run() {
			while (true) {
				try {
					Thread.sleep(sleep);
				} catch (InterruptedException ie) {
					// Nothing to do
				}
				if (!Configuration.this.listeners.isEmpty()) {
					Configuration.this.check();
				}
			}
		}
	}
}
