/*
 * The SmartConfig Library
 * Copyright (C) 2004-2006
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * For further informations on the SmartConfig Library please visit
 *
 *                        http://smartconfig.sourceforge.net
 */
package net.smartlab.config;

import java.util.Collection;

/**
 * @author rlogiacco
 */
public class Reference extends Element {

	/**
	 * The name of the element attribute defining a reference to another
	 * element.
	 */
	protected final static String REFERENCE = "refid";

	/**
	 * The name of the element attribute defining a valid reference target.
	 */
	protected final static String IDENTIFIER = "id";

	/**
	 * The referenced node. It will be <code>null</code> until resolved.
	 */
	protected Node node;

	/**
	 * The unique identifier for this reference.
	 * 
	 * @uml.property name="id"
	 */
	protected String id;


	/**
	 * Creates a reference contained by the given parent node with a specified
	 * name and unique identifier.
	 * 
	 * @param parent the node containing the element.
	 * @param name the element name.
	 * @param id the element identifier.
	 */
	protected Reference(Node parent, String name, String id) {
		super(parent, name);
		this.id = id;
	}

	/**
	 * @see net.smartlab.config.Element#getId()
	 * @uml.property name="id"
	 */
	public String getId() {
		return id;
	}

	/**
	 * @see net.smartlab.config.Element#getElements()
	 */
	public Collection getElements() throws ConfigurationException {
		if (node == null) {
			node = parent.resolve(name, id);
		}
		try {
			return node.getElements();
		} catch (NullPointerException npe) {
			throw new ConfigurationException("Undefined reference to node " + id);
		}
	}

	/**
	 * @see net.smartlab.config.Element#getElements(java.lang.String)
	 */
	public Collection getElements(String name) throws ConfigurationException {
		if (node == null) {
			node = parent.resolve(this.name, id);
		}
		try {
			return node.getElements(name);
		} catch (NullPointerException npe) {
			throw new ConfigurationException("Undefined reference to node " + id);
		}
	}

	/**
	 * @see net.smartlab.config.Element#getElement(java.lang.String,
	 *      java.lang.String, java.lang.String)
	 */
	public Element getElement(String name, String attribute, String value) throws ConfigurationException {
		if (node == null) {
			node = parent.resolve(this.name, id);
		}
		try {
			return node.getElement(name, attribute, value);
		} catch (NullPointerException npe) {
			throw new ConfigurationException("Undefined reference to node " + id);
		}
	}

	/**
	 * @see net.smartlab.config.Element#getAttributes()
	 */
	public Collection getAttributes() throws ConfigurationException {
		if (node == null) {
			node = parent.resolve(name, id);
		}
		try {
			return node.getAttributes();
		} catch (NullPointerException npe) {
			throw new ConfigurationException("Undefined reference to node " + id);
		}
	}

	/**
	 * @see net.smartlab.config.Element#getAttributeNames()
	 */
	public Collection getAttributeNames() throws ConfigurationException {
		if (node == null) {
			node = parent.resolve(name, id);
		}
		try {
			return node.getAttributeNames();
		} catch (NullPointerException npe) {
			throw new ConfigurationException("Undefined reference to node " + id);
		}
	}

	/**
	 * @see net.smartlab.config.Element#getAttribute(java.lang.String)
	 */
	public String getAttribute(String name) throws ConfigurationException {
		if (name.equalsIgnoreCase(REFERENCE)) {
			return this.id;
		}
		if (node == null) {
			node = parent.resolve(this.name, id);
		}
		try {
			return node.getAttribute(name);
		} catch (NullPointerException npe) {
			throw new ConfigurationException("Undefined reference to node " + id);
		}
	}

	/**
	 * @see net.smartlab.config.Element#getContent()
	 */
	public String getContent() throws ConfigurationException {
		if (node == null) {
			node = parent.resolve(name, id);
		}
		try {
			return node.content;
		} catch (NullPointerException npe) {
			throw new ConfigurationException("Undefined reference to node " + id);
		}
	}

	/**
	 * @see net.smartlab.config.Element#resolve()
	 */
	public void resolve() throws ConfigurationException {
		if (node == null) {
			node = parent.resolve(name, id);
		}
	}
}
